/* ---------------------------------------------------------------------------------
 * Rseau en anneau  jeton (token ring) : mise en place de l'anneau
 * Jean-Paul Rigault 
 * ---------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>

/* Dfinition globales
 * -------------------
 */

#define STATION_EXE "./station.exe"	/* nom de l'xcutable d'une station */

#define MAX_STATIONS 10			/* Nombre maximum de stations */

int n_stations;				/* Nombre rel de stations */
int fd_pipes[MAX_STATIONS][2];		/* Descripteurs de tous les pipes */

/* Dclaration en avant */
void station(int i);

/* Programme principal
 * -------------------
 *
 * On cre les pipes, puis les n processus fils correspondant aux stations. La
 * fonction station(i), invoque par chaque fils, tablit les connexions et
 * excute le programme station.exe.
 */
int main(int argc, char *argv[])  
{ 
    int i; 

    /* Extraction du nomnre de stations pass en argument */
    if (argc <= 1)
    {
	fprintf(stderr, "Usage : ring.exe n_stations\n");
	exit(1);
    }

    n_stations = atoi(argv[1]);
    if (n_stations <= 2 || n_stations >= MAX_STATIONS)
    {
	fprintf(stderr, "Vous voulez %d stations ? Soyez raisonnable! \n"
                "\t(Nombre maximum possible de stations : %d)\n", 
		n_stations, MAX_STATIONS );
	exit(1);
    }

    /* Creation de tous les pipes */
    for (i = 0; i < n_stations; i++)
    {
        if (pipe(fd_pipes[i]) < 0) perror("cration des tubes");
    }

    /* Creation des processus des stations
     * -----------------------------------
     * Le processus Si lit sur le pipe i et crit sur le pipe i+1. La structure
     * est circulaire: les calculs d'indice ont donc lieu modulo n_stations
     * (c'est pourquoi il vaut mieux viter les valeurs ngatives de i)
     */
    for (i = 0; i < n_stations; ++i)
    {
	switch(fork())
	{
	case -1:
	    perror("fork");
	    break;
	case 0:
	    /* Le fils appelle la mise en place de la station i */
	    station(i);
	    break;
	default:
	    /* Le pre passe  l'itration suivante */
	    break;
	}
    }

    /* Le pre se met en attente de tous ses fils */
    for (i = 0; i < n_stations; i++) wait(NULL);

    /* Fin du programme */
    exit(0);
}

/* Mise en place d'une station
 * ---------------------------
 */
void station(int i)
{
    int j;
    char ctab[5]; 
    char *arg[] = {STATION_EXE, NULL, NULL}; /* Argument de execvp */

    /* On redirige entre et sortie standard */
    dup2(fd_pipes[i][0], 0);
    dup2(fd_pipes[(i + 1) % n_stations][1], 1);

    /* On peut maintenant fermer tous les descripteurs de pipes */
    for (j = 0; j < n_stations; j++)
    {
	close(fd_pipes[j][0]);
	close(fd_pipes[j][1]);
    }

    /* On excute la station en passant i en argument */
    sprintf(ctab, "%d", i); /* encodage de i sous forme d'une chane */ 
    arg[1] = ctab;
    execvp(STATION_EXE, arg);
    perror("excution de la station");
    exit(1); /* surtout pas return ! */
}
