/* ---------------------------------------------------------------------------------
 * Tubes du shell
 * Jean-Paul Rigault 
 * ---------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <wait.h>

/* Fermeture des deux descripteurs d'un pipe */
static void close_pipe(int p[2])
{
    close(p[0]);
    close(p[1]);
}

/* Programme principal
 * -------------------
 *
 * NOTE : il est indispensable que chaque processus ferme les descripteurs de
 * pipes qu'il n'utilise pas (cad, ici, tous, une fois les redirections
 * ralises).
 *
 * En effet, les processus grep et wc se terminent sur une fin de fichier (EOF)
 * en entre. Pour que EOF soit dtecter sur un pipe, il faut qu'il n'y ait
 * plus AUCUN processus qui ait ce pipe ouvert en criture.
 *
 * L'exprience est simple  faire : commentez le corps de la fonction
 * close_pipe() ci-dessus, recompilez, et excutez. Votre excution va se
 * bloquer, et ps (dans une autre fentre vous montrera que vos processus grep
 * et wc sont toujours vivants.
 */
int main()
{
    int p1[2];
    int p2[2];
    int i;

    switch (fork()) 
    {
    case -1:
        perror("fork");
        exit(1);

    case 0: /* Le fils excute l'ensemble du pipe */
        
        /* On cre les deux pipes */
        if (pipe(p1) < 0 || pipe(p2) < 0)
        {
            perror("pipes");
            exit(1);
        }

        /* On cre 3 fils successifs. Pour allger le code, on ne tient pas
        * compte des erreurs possibles lors du fork() (d'ailleurs assez
        * rares !)
        */

        /* Premier fils : il excute ps */
        if (fork() == 0) 
        {
            dup2(p1[1], 1);	/* redirection de la sortie standard */
            close_pipe(p1);	/* ces descripteurs sont maintenant inutiles  */            close_pipe(p2);
            execlp("ps", "ps", NULL);
            perror("exit ps");
            exit(1);
        }

        /* Deuxime fils : il excute grep */

        if (fork() == 0) 
        {
            dup2(p1[0], 0);	/* redirection de l'entre standard */
            dup2(p2[1], 1);	/* redirection de la sortie standard */
            close_pipe(p1);	/* ces descripteurs sont maintenant inutiles */             close_pipe(p2);
            execlp("grep", "grep", "emacs", NULL);
            perror("exit grep");
            exit(1);
        }

        /* Troisime fils : il excute wc */

        if (fork() == 0) 
        {
            dup2(p2[0], 0);	/* redirection de l'entre standard */
            close_pipe(p1);	/* ces descripteurs sont maintenant inutiles  */
            close_pipe(p2);
            execlp("wc", "wc", "-l", NULL);
            perror("exit wc");
            exit(1);
        }       
     
    default: /* Le pre attend tous ses fils (3) */

        close_pipe(p1);	/* ces descripteurs sont maintenant inutiles */
        close_pipe(p2);
        for (i = 0; i < 3; ++i) wait(0);
    }

    return 0;
}
        
